;;; **********************************************************************
;;; Copyright (c) 89-93, 94 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

#+(or cltl2 lispworks clisp) 
(progn
  (defpackage :COMMON-MUSIC
    (:use #-(and clisp pcl) :COMMON-LISP #+(and clisp pcl) :LISP
         :TL
          #+(or pcl mcl lispworks clisp) :WALKER
          #+pcl :PCL
          #+(or (and excl cltl2) lispworks) :CLOS
          #+clm :CLM
    ) 
    #+(and clisp pcl) (:import-from "PCL" "PRINTING-RANDOM-THING")
    (:nicknames :cm)
    #+clisp (:shadow "SHELL"))
  (in-package :common-music))

#-(or cltl2 lispworks clisp) 
(in-package :common-music
            :use (list (find-package ':lisp)
                       (find-package ':walker)
                 #+pcl (find-package ':pcl)  ; aclpc has clos but not walk
                       (find-package ':tl)
                 #+clm (find-package ':clm))
            :nicknames '(:cm))

(defvar *common-music-directory* 
  #+excl (directory-namestring excl::*source-pathname*)
  #+mcl  (directory-namestring ccl:*loading-file-source-file*)
  #+akcl (directory-namestring si:*load-pathname*)
  #+lispworks (directory-namestring *load-pathname*)
  ;; this horrible hack is our last ditch effort at finding cm's source
  ;; directory. works if the user is loading via build.lisp, the normal case.
  #+clisp (directory-namestring *load-pathname*)
  #-(or excl akcl mcl lispworks clisp)
  (let ((dir (find-symbol "CM-DIRECTORY"
                          (or (find-package :user)
                              (find-package :common-lisp-user)))))
    (if dir 
        (symbol-value dir)
      (error "~%The system cannot automatically determine the source directory for Common Music in this Lisp port.  Edit the file common-music.lisp and set the variable *common-music-directory* to the appropriate directory namestring."))))
  
(defvar *common-music-package* (find-package :common-music))
(defvar *common-music-version* "Common Music 11-Oct-94")

#+cmn
(import '(cmn::size cmn::metronome))

#+(and excl cltl2)
(import '(clos::walk-form clos::variable-lexical-p clos::variable-special-p
	  clos::variable-globally-special-p))
	  
#+aclpc
(import '(walker::walk-form walker::variable-lexical-p 
          walker::variable-special-p walker::variable-globally-special-p))
	  
#+pcl
(eval-when (compile load eval)
  #-clisp (import '(pcl::printing-random-thing))
  (cond ((string= pcl::*pcl-system-date* "March 92 PCL (2c)")
         (import '(pcl::class-prototype pcl::class-slots
                   pcl::allocate-instance pcl::standard-class 
                   pcl::validate-superclass pcl::class-precedence-list	    
                   pcl::class-direct-superclasses pcl::class-direct-subclasses 
                   pcl::class-direct-slots pcl::finalize-inheritance	    
                   pcl::class-finalized-p pcl::class-direct-slots	    
                   pcl::slot-value-using-class pcl::slot-boundp-using-class
                   pcl::slot-makunbound-using-class pcl::slot-definition-name
                   pcl::slot-definition-readers	pcl::slot-definition-writers
                   pcl::slot-definition-initargs pcl::slot-definition-initform
                   pcl::slot-definition-initfunction 
                   pcl::slot-definition-allocation
                   pcl::slot-definition-type)
                 (find-package :cm)))
        ((string= pcl::*pcl-system-date* "September 16 92 PCL (f)")
         nil)
        (t
         (error "~&Common Music has only been tested in March 92 PCL (2c) and September 16 92 PCL (f). It may work in your version of PCL, but then again it may not.  The easiest thing to do is to ftp March 92 PCL from guido.zkm.de and install that.  Otherwise, edit the files: common-music.lisp and stella/pkg.lisp to get rid of this error message code, and then figure out the import statements that you need for your version of pcl.")))
)

#+mcl
(progn
  (import '(ccl:class-direct-superclasses ccl:class-direct-subclasses 
            ccl:class-prototype ccl:class-precedence-list
            ccl:slot-definition-name))
  (defun class-slots (class) (ccl:class-instance-slots class))
  (defun class-direct-slots (class) (ccl:class-direct-instance-slots class))
  (defun slot-definition-initargs (slot) (third slot))
  (defun slot-definition-initform (slot) 
    (and (second slot) 
         (if (functionp (second slot))
             (funcall (second slot))
           (if (consp (second slot))
               (car (second slot))
            nil))))
  (defun slot-definition-readers (slot)
    (declare (ignore slot))
    (error "slot-definition-readers undefined in mcl"))
  (defun slot-definition-writers (slot) 
    (declare (ignore slot))
    (error "slot-definition-writers undefined in mcl"))
  (defmacro slot-value-using-class (class instance slot)
    (declare (ignore class))
    `(slot-value ,instance (ccl:slot-definition-name ,slot)))
  (defmacro slot-boundp-using-class (class instance slot)
    (declare (ignore class))
    `(slot-boundp ,instance (ccl:slot-definition-name ,slot)))
  (defmacro slot-makunbound-using-class (class instance slot)
    (declare (ignore class))
    `(slot-makunbound ,instance (ccl:slot-definition-name ,slot)))
)

#+(and clisp (not pcl))
(progn
  (defun allocate-instance (class)
    (clos::std-allocate-instance class))
  (defun class-slots (class)
    (clos::class-slots class))
  (defun class-precedence-list (class)
    (clos::class-precedence-list class))
  (defun class-direct-superclasses (class)
    (clos::class-direct-superclasses class))     
  (defun slot-definition-name (slot)
    (clos::slotdef-name slot))
  (defun slot-definition-initargs (slot)
    (clos::slotdef-initargs slot))
  (defun slot-definition-initform (slot)
    (let ((? (clos::slotdef-initer slot)))
      (and ? (if (car ?) (funcall (car ?)) (cdr ?)))))
  (defun slot-definition-readers (slot)
    (declare (ignore slot))
    (error "slot-definition-readers undefined in CLISP"))
  (defun slot-definition-writers (slot) 
    (declare (ignore slot))
    (error "slot-definition-writers undefined in CLISP"))
  (defmacro slot-value-using-class (class object slot)
    (declare (ignore class))
    `(slot-value ,object (slot-definition-name ,slot)))
  (defun slot-boundp-using-class (class object slot)
    (declare (ignore class))
    (slot-boundp object (slot-definition-name slot)))
  (defun slot-makunbound-using-class (class object slot)
    (declare (ignore class))
    (slot-makunbound object (slot-definition-name slot)))
  (defun change-class (object)
    (declare (ignore object))
    (error "CLISP's CLOS does not implement change-class."))
  (defvar %clisp-prototypes% (make-hash-table))
  (defun class-prototype (class)
    (or (gethash class %clisp-prototypes% )
        (setf (gethash class %clisp-prototypes% )
              (clos::std-allocate-instance class))))
)

#+(and clisp dos)
(progn
  (defun user-homedir-pathname () (pathname "C:\\"))
)

(export '(*common-music-directory*

          ;; utilities.lisp
          quotify *common-music-readtable* save-cm defprop
          +source-type+ +binary-type+ +midi-type+ +text-type+
          +stella-type+ +sound-type+ +killed+ +ending+ +resting+
          +chording+ +unqueued+ +removed+ +normal-mask+

          ;; item-streams.lisp
          item +end-of-stream-token+ +eop+ +eod+ read-items doitems
          last-item last-state expr period

          ;; rhythm-streams.lisp
          *standard-tempo* in-tempo rhythm

          ;; functions.lisp
          *coordinates-are-x-y-pairs* interpolation interpl
          lookup function-value

          ;; scales.lisp
          general-scale equal-tempered-scale gapped-scale find-scale
          list-all-scales cents centify *standard-scale* 
          *respect-note-spelling*
          in-scale standard-chromatic-scale *standard-chromatic-scale*
          make-scale scale-pitch scale-degree scale-note octave-and-interval
          note pitch degree restp scale> scale< scale= scale<= scale>=
          scale/= scale- defscale
	  ; scale degrees of standard scale
          C00 CN00 CS00 DF00 D00 DN00 DS00 EF00 E00 EN00 FF00 F00 FN00 ES00
          FS00 GF00 G00 GN00 GS00 AF00 A00 AN00 AS00 BF00 B00 BN00 CF0 C0 CN0
          BS00 CS0 DF0 D0 DN0 DS0 EF0 E0 EN0 FF0 F0 FN0 ES0 FS0 GF0 G0 GN0 GS0
          AF0 A0 AN0 AS0 BF0 B0 BN0 CF1 C1 CN1 BS0 CS1 DF1 D1 DN1 DS1 EF1 E1
          EN1 FF1 F1 FN1 ES1 FS1 GF1 G1 GN1 GS1 AF1 A1 AN1 AS1 BF1 B1 BN1 CF2
          C2 CN2 BS1 CS2 DF2 D2 DN2 DS2 EF2 E2 EN2 FF2 F2 FN2 ES2 FS2 GF2 G2
          GN2 GS2 AF2 A2 AN2 AS2 BF2 B2 BN2 CF3 C3 CN3 BS2 CS3 DF3 D3 DN3 DS3
          EF3 E3 EN3 FF3 F3 FN3 ES3 FS3 GF3 G3 GN3 GS3 AF3 A3 AN3 AS3 BF3 B3
          BN3 CF4 C4 CN4 BS3 CS4 DF4 D4 DN4 DS4 EF4 E4 EN4 FF4 F4 FN4 ES4 FS4
          GF4 G4 GN4 GS4 AF4 A4 AN4 AS4 BF4 B4 BN4 CF5 C5 CN5 BS4 CS5 DF5 D5
          DN5 DS5 EF5 E5 EN5 FF5 F5 FN5 ES5 FS5 GF5 G5 GN5 GS5 AF5 A5 AN5 AS5
          BF5 B5 BN5 CF6 C6 CN6 BS5 CS6 DF6 D6 DN6 DS6 EF6 E6 EN6 FF6 F6 FN6
          ES6 FS6 GF6 G6 GN6 GS6 AF6 A6 AN6 AS6 BF6 B6 BN6 CF7 C7 CN7 BS6 CS7
          DF7 D7 DN7 DS7 EF7 E7 EN7 FF7 F7 FN7 ES7 FS7 GF7 G7 GN7 GS7 AF7 A7
          AN7 AS7 BF7 B7 BN7 CF8 C8 CN8 BS7 CS8 DF8 D8 DN8 DS8 EF8 E8 EN8 FF8
          F8 FN8 ES8 FS8 GF8 G8 GN8 GS8 AF8 A8 AN8 AS8 BF8 B8 BN8
          R A B C D E F G AN BN CN DN EN FN GN AF BF CF DF EF FF GF AS BS CS
          DS ES FS GS

          ;; amplitude-streams.lisp
          amplitude *amplitude-minimum* *amplitude-maximum* *amplitude-power*
          niente pppp ppp pp p mp mf f ff fff ffff

	  ;; multiple-items.lisp
          defmultiple-item defmultiple-item-parser defmultiple-item-filter
          multiple-item-bind

          ;; music-utilities.lisp
          transpose invert between header pick pickl

          ;; constructors.lisp
          make-item-stream items notes pitches degrees rhythms intervals 
          steps amplitude amplitudes series numbers chord motive mirror
          repeat idsel crescendo diminuendo tempo changes make-chord expr fn
          ; patterns
          cycle sequence accumulation palindrome random heap graph
          function rotation
          ; options
          above args below by change counting downto for elided forming from in
          initially-from linked-to loudest modulus multiple named of on power
          previous props returning softest tempo to traversing using with
          ; option constants
          id weight min max yes no start end p i r ri prime inversion step
          width retrograde retrograde-inversion idsel depth breadth periods
          values
          ; item stream classes
          item-stream cyclic-item-stream random-item-stream heap-item-stream
          palindromic-item-stream graph-item-stream sequential-item-stream
          functional-item-stream rotational-item-stream
          accumulating-item-stream cyclic-note-stream random-note-stream
          heap-note-stream palindromic-note-stream graph-note-stream
          sequential-note-stream functional-note-stream rotational-note-stream 
          accumulating-note-stream cyclic-pitch-stream random-pitch-stream
          heap-pitch-stream palindromic-pitch-stream graph-pitch-stream
          sequential-pitch-stream functional-pitch-stream 
          rotational-pitch-stream accumulating-pitch-stream
          cyclic-degree-stream random-degree-stream heap-degree-stream
          palindromic-degree-stream graph-degree-stream
          sequential-degree-stream functional-degree-stream
          rotational-degree-stream accumulating-degree-stream
          cyclic-interval-stream random-interval-stream heap-interval-stream
          palindromic-interval-stream graph-interval-stream 
          sequential-interval-stream functional-interval-stream
          rotational-interval-stream  accumulating-interval-stream
          cyclic-step-stream random-step-stream heap-step-stream
          palindromic-step-stream graph-step-stream sequential-step-stream
          functional-step-stream rotational-step-stream
          accumulating-step-stream cyclic-rhythm-stream random-rhythm-stream
          heap-rhythm-stream palindromic-rhythm-stream graph-rhythm-stream
          sequential-rhythm-stream functional-rhythm-stream
          rotational-rhythm-stream accumulating-rhythm-stream
          cyclic-amplitude-stream random-amplitude-stream heap-amplitude-stream
          palindromic-amplitude-stream graph-amplitude-stream
          sequential-amplitude-stream functional-amplitude-stream
          rotational-amplitude-stream accumulating-amplitude-stream
          series-stream linear-number-stream random-number-stream
          chord-stream

          ;; clm.lisp
          *default-scorefile-with-sound*

          ;; messages.lisp
          make-channel-message channel-message-p channel-message-status
          channel-message-channel channel-message-data1 channel-message-data2
          make-note-on note-on-p note-on-channel note-on-key note-on-velocity 
          make-note-off note-off-p note-off-channel note-off-key
          note-off-velocity make-key-pressure key-pressure-p
          key-pressure-channel key-pressure-key key-pressure-pressure
          make-control-change control-change-p control-change-channel
          control-change-control control-change-change make-program-change
          program-change-p program-change-channel program-change-program
          make-channel-pressure channel-pressure-p channel-pressure-channel
          channel-pressure-pressure make-pitch-bend pitch-bend-p
          pitch-bend-channel pitch-bend-lsb pitch-bend-msb
          make-channel-mode-message channel-mode-p channel-mode-channel
          channel-mode-control channel-mode-change make-system-message
          system-message-p system-message-status system-message-data1
          system-message-data2 make-song-position song-position-p
          song-position-lsb song-position-msb make-song-select
          song-select-p song-select-song make-tune-request tune-request-p
          make-sysex-message sysex-message-p sysex-message-id 
          make-meta-message meta-message-p meta-message-type
          make-time-signature time-signature-p make-tempo-change
          tempo-change-p make-eot eot-p make-midi-data midi-data-p
          midi-data-size midi-data-data1 midi-data-data2
          midi-data-data3 make-midi-message

          ;; midi.lisp
          *default-midi-pathname* midi-open midi-open-p midi-port-reference-p
          ask-port midi-close midi-hush midi-print-message midi-write-message
          midi-read-messages midi-get-time midi-set-time midi-set-quanta-size
          quanta-time real-time midi-stop-time midi-start-time
          *midi-read-hook* with-midi-open midi-receive-messages
  	  
          ;; midifile.lisp
          midifile-play midifile-print midifile-map midifile-parse
          write-message write-midi-file-header read-midi-file-header
          midifile-to-vector
	  ))

(pushnew ':common-music *features*)
