;;; -*- syntax: common-lisp; package: clm; base: 10; mode: lisp -*-
;;;
;;; support for simple list ops in run
;;;
;;; these make no real effort to protect the user from himself or provide every possible option.
;;; Also, unless otherwise stated, none of these support setf 
;;; Currently implemented are first through tenth with the equivalent car form, 
;;;     list-length, pop, push, member, fill, delete, nsubstitute, count

(in-package :clm)

(def-dsp-fun 'FIRST #'(lambda (var x) (package-op '<first_c> var x 'list)))
(def-dsp-fun 'CAR #'(lambda (var x) (package-op '<first_c> var x 'list)))
(defmacro <first_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]];~A/* (first ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'SECOND #'(lambda (var x) (package-op '<second_c> var x 'list)))
(def-dsp-fun 'CADR #'(lambda (var x) (package-op '<second_c> var x 'list)))
(defmacro <second_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+1];~A/* (second ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'THIRD #'(lambda (var x) (package-op '<third_c> var x 'list)))
(def-dsp-fun 'CADDR #'(lambda (var x) (package-op '<third_c> var x 'list)))
(defmacro <third_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+2];~A/* (third ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'FOURTH #'(lambda (var x) (package-op '<fourth_c> var x 'list)))
(def-dsp-fun 'CADDDR #'(lambda (var x) (package-op '<fourth_c> var x 'list)))
(defmacro <fourth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+3];~A/* (fourth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'FIFTH #'(lambda (var x) (package-op '<fifth_c> var x 'list)))
(defmacro <fifth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+4];~A/* (fifth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'SIXTH #'(lambda (var x) (package-op '<sixth_c> var x 'list)))
(defmacro <sixth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+5];~A/* (sixth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'SEVENTH #'(lambda (var x) (package-op '<seventh_c> var x 'list)))
(defmacro <seventh_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+6];~A/* (seventh ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'EIGHTH #'(lambda (var x) (package-op '<eighth_c> var x 'list)))
(defmacro <eighth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+7];~A/* (eighth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'NINTH #'(lambda (var x) (package-op '<ninth_c> var x 'list)))
(defmacro <ninth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+8];~A/* (ninth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))

(def-dsp-fun 'TENTH #'(lambda (var x) (package-op '<tenth_c> var x 'list)))
(defmacro <tenth_c> (result-name lst) 
  (format *c-file* "  ~A = _datar_[~A[~D]+9];~A/* (tenth ~A) */~%" (lc result-name) (lc lst) aref_block cr_white_space lst))


(def-dsp-fun 'LIST-LENGTH #'(lambda (var x) (package-op '<list-length_c> var x 'list)))
(defmacro <list-length_c> (result-name lst) 
  (format *c-file* "  ~A = ~A[~D];~A/* (list-length ~A) */~%" (lc result-name) (lc lst) aref_size cr_white_space lst))


;;; pop and push are intended for small stack ops (nothing fancy!)
(def-dsp-fun 'PUSH #'(lambda (var x)
		       (chkv (third x) 'list)
		       (push (third x) true-list-vars)
		       (package-op '<push_c> var x)))

(defmacro <push_c> (result-name x lst)
  (declare (ignore result-name))	;this is not a "full" push implementation
  (let ((lca (lc lst))
        (lcx (lc x)))
    (format *c-file* "  {int _size_,_i_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (push ~A ~A) */~%"
          lca aref_size lca aref_block cr_white_space lst x)
    (format *c-file* "   for (_i_=_size_;_i_>0;_i_--) _arr_[_i_]=_arr_[_i_-1]; _arr_[0]=~A; ~A[~D]++;}~%" lcx lca aref_size)))
;;; would be nicer to have the array stored so that the head space is free -- pops and pushes 
;;; would then be much simpler.

(def-dsp-fun 'POP #'(lambda (var x)
		      (push (second x) true-list-vars)
		      (package-op '<pop_c> var x 'list)))

(defmacro <pop_c> (result-name lst)
  (let ((lcr (lc result-name))
	(lca (lc lst)))
    (format *c-file* "  {int _size_,_i_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (pop ~A) */~%"
          lca aref_size lca aref_block cr_white_space lst)
    (format *c-file* "   ~A = _arr_[0]; ~A[~D]--; for (_i_=0;_i_<_size_;_i_++) _arr_[_i_]=_arr_[_i_+1];}~%" lcr lca aref_size)))

    
(def-dsp-fun 'MEMBER #'(lambda (var x)
			 (chkv (third x) 'list)
			 (package-op '<member_c> var x)))

(defmacro <member_c> (result-name x lst)
  (let ((lcr (lc result-name))
	(lca (lc lst))
	(lcx (lc x)))
    (format *c-file* "  {int _size_,_i_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (member ~A ~A) */~%"
          lca aref_size lca aref_block cr_white_space x lst)
    (format *c-file* "   ~A = 0; ~A = 0.0; _i_ = 0; while ((_i_<_size_) && (~A == 0)) {if (_arr_[_i_] == ~A) ~A = 1; ~A = ~A; _i_++;}}~%"
	    c-test-Z lcr c-test-Z lcx c-test-Z lcr lcx)))


(def-dsp-fun 'FILL #'(lambda (var x) (package-op '<fill_c> var x 'list)))

(defmacro <fill_c> (result-name lst x)
  (declare (ignore result-name))
  (let ((lca (lc lst))
	(lcx (lc x)))
    (format *c-file* "  {int _size_,_i_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (fill ~A ~A) */~%"
          lca aref_size lca aref_block cr_white_space lst x)
    (format *c-file* "   for (_i_=0;_i_<_size_;_i_++) _arr_[_i_]=~A;}~%" lcx)))


(def-dsp-fun 'DELETE #'(lambda (var x) 
			 (chkv (third x) 'list)
			 (package-op '<delete_c> var x)))

(defmacro <delete_c> (result-name x lst)
  (declare (ignore result-name))
  (let ((lca (lc lst))
	(lcx (lc x)))
    (format *c-file* "  {int _size_,_i_,_j_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (delete ~A ~A) */~%"
          lca aref_size lca aref_block cr_white_space x lst)
    (format *c-file* "   for (_i_=0,_j_=0;_i_<_size_;_i_++) {if (_arr_[_i_] != ~A) {if (_i_ != _j_) _arr_[_j_] = _arr_[_i_]; _j_++;}}~%" lcx)
    (format *c-file* "   ~A[~D] = _j_+1;}~%")))


(def-dsp-fun 'NSUBSTITUTE #'(lambda (var x) 
			      (chkv (fourth x) 'list)
			      (package-op '<nsubstitute_c> var x)))

(defmacro <nsubstitute_c> (result-name x y lst)
  (declare (ignore result-name))
  (let ((lca (lc lst))
	(lcx (lc x))
	(lcy (lc y)))
    (format *c-file* "  {int _size_,_i_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; ~A/* (nsubstitute ~A ~A ~A) */~%"
          lca aref_size lca aref_block cr_white_space x y lst)
    (format *c-file* "   for (_i_=0;_i_<_size_;_i_++) {if (_arr_[_i_] == ~A) _arr_[_i_]=~A;}}~%" lcy lcx)))


(def-dsp-fun 'COUNT #'(lambda (var x)
			 (chkv (third x) 'list)
			 (package-op '<count_c> var x)))

(defmacro <count_c> (result-name x lst)
  (let ((lcr (lc result-name))
	(lca (lc lst))
	(lcx (lc x)))
    (format *c-file* "  {int _size_,_i_,_count_; float *_arr_; _size_ = ~A[~D]; _arr_ = _datar_+~A[~D]; _count_=0;~A/* (count ~A ~A) */~%"
	    lca aref_size lca aref_block cr_white_space x lst)
    (format *c-file* "   for (_i_=0;_i_<_size_;_i_++) {if (_arr_[_i_] == ~A) _count_++;} ~A = ~A_count_;}~%"
	    lcx lcr (if (int-p result-name) "" "(int)"))))


