;;; -*- Syntax: Common-Lisp; Package: clm; Base: 10; Mode: Lisp -*-

(in-package :clm)
 
(defun mac-ram-size ()
  (%stack-block ((m 4)) 
    (#_gestalt "ram " m)
    (%get-unsigned-long m)))

(defun mac-ram-in-MB ()
  (floor (mac-ram-size) (expt 2 20)))

(defconstant kMachineNameStrID -16395)

(defun mac-machine-type ()
  (%stack-block ((m 4)) 
    (#_gestalt "mach" m)
    (let ((n (%get-unsigned-long m)))
      (ccl::get-ind-string kMachineNameStrID n))))

(defun mac-os-version ()
  ;; returns the digits in successive 8-bit bytes as (for example) 1808 = #x710 = system 7.1
  (%stack-block ((m 4)) 
    (#_gestalt "sysv" m)
    (%get-unsigned-long m)))

(defun mac-os-name ()
  (let ((n (format nil "~X" (mac-os-version))))
    (concatenate 'string (subseq n 0 1) "." (subseq n 1 2) "." (subseq n 2 3))))

(defun mac-processor ()
  (%stack-block ((m 4)) 
    (#_gestalt "proc" m)
    (let ((n (%get-unsigned-long m)))
      (case n
	(1 "68000")
	(2 "68010")
	(3 "68020")
	(4 "68030")
	(5 "68040")
	(otherwise (format nil "unknown: ~D" n))))))

(defun mac-sound-attributes ()
  (%stack-block ((m 4)) 
    (#_gestalt "proc" m)
    (let ((n (%get-unsigned-long m)))
      n)))
;;; a table of attributes:
;;; 0=stereo, 1=mixed internally, 3=new sound manager, 4=sound-in built-in, 5=sound-in from somewhere,
;;; 6=can play and record, 7=has 16-bit, 8=can record stereo, 9=line level input, 10=play from disk ok,
;;; 11=not just mono, 12=16-bit supported

(defun mac-slot-locations ()
  (%stack-block ((m 4)) 
    (#_gestalt "sltc" m)
    (let ((n (%get-unsigned-long m)))
      n)))
;;; this is a table -- bit on = that slot is available in the hardware

(defun mcl-get-string (num) 
  (#_GetString num))

(defun get-volume ()			;apparently a number between 0 and 7
  (traps::getsoundvol))			;from MCL 2.0:library:interfaces.lisp

(defun set-volume (vol)
  (traps::setsoundvol vol))


;;; the dac routine supplied by Tobias Kunze after I gave up.

(defun start-dac (file &optional ignore-error start end)
  (declare (ignore ignore-error))
  (let* ((fil (open file :direction :input :element-type 'unsigned-byte))
	 (ref (ccl::rref (ccl::uvref (cdr (ccl::column.fblock fil)) 1) :cinfopbrec.iofrefnum))
	 (tend (if (and end (> end 0.0)) end 
		 (progn 
		   (c-read-header file) 
		   (float (/ (c-snd-header-data-size) 
			     (* (c-snd-header-srate) (c-snd-header-chans) 2)))))))
    (ccl::rlet ((selection :audioselection
			   :unittype #$unittypeseconds
			   :selstart (if start (+ (ash (floor start) 16) (round (* (rem start 1) 65536))) 0)
			   :selend (+ (ash (floor tend) 16) (round (* (rem tend 1) 65536)))))
      (#_SndStartFilePlay (ccl::%null-ptr) ref 0 #x8000 (ccl::%null-ptr) selection (ccl::%null-ptr) nil))
    (close fil)
    0))


(defun make-handle-i (arr)
  (let ((size (length arr))
	(hd (#_NewHandle :check-error 0)))
    (unless (ccl:handlep hd)
      (error "#_newHandle failed with ~A." hd))
    (#_SetHandleSize :check-error hd (* size 4))
    (ccl:with-dereferenced-handles ((p hd))
      (dotimes (i size)
	(ccl:%put-long p (aref arr i) (* i 4))))
    hd))

(defun free-handle-i (hd arr)
  (unless (ccl:handlep hd)
    (error "~S is not a valid handle." hd))
  (let ((size (floor (#_GetHandleSize :check-error hd) 4)))
    (ccl:with-dereferenced-handles ((p hd))
      (dotimes (i size)
	(setf (aref arr i) (ccl:%get-long p (* i 4))))))
  (#_DisposHandle :check-error hd))

(defun make-handle-f (arr)
  (let ((size (length arr))
	(hd (#_NewHandle :check-error 0)))
    (unless (ccl:handlep hd)
      (error "#_newHandle failed with ~A." hd))
    (#_SetHandleSize :check-error hd (* size 8))
    (ccl:with-dereferenced-handles ((p hd))
      (dotimes (i size)
	(multiple-value-bind (int frac) (floor (aref arr i))
	  (ccl:%put-long p int (* i 8))
	  (ccl:%put-long p (floor (* 1073741824 frac)) (+ 4 (* i 8))))))
    hd))

(defun free-handle-f (hd arr)
  (declare (ignore arr))
  (unless (ccl:handlep hd)
    (error "~S is not a valid handle." hd))
  (#_DisposHandle :check-error hd))


			  
