;;; **********************************************************************
;;; Copyright (c) 89, 90, 91, 92, 93 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; send to: hkt@zkm.de
;;; **********************************************************************
;;; Cmix syntax by Charles Baker 5/94 - based completely on hkt's code!
;;; Most standard cmix calls implemented in here, BUT
;;; it is the user's responsability to make sure the "orchestra" is
;;; a cmix executable with all the calls linked in!
;;; This is the "orchestra" slot in the file.
;;; I also try to signal an error if you try to open too many cmix i/o channels
;;; in the file.
;;; You can specify "outskip" slot in  mix and trans commands directly 
;;; (as in cmix), or you can use "rhythm" to increment time, as in usual 
;;; CommonMusic output...up to you! (See write event for mix and trans objects)
(in-package :stella)

(defsyntax cmix 
  :pathname (pathname "test.mix")
  :stream-types '((cmix-score-file "mix"))
  :nicknames '(cmix))

(defclass cmix-score-file (event-file header-mixin)
  ((orchestra :initarg orchestra :initform "") ;; what compiled executable to use for this score file.
  (openfiles :initarg openfiles :initform ()) ;; list of open i/o files  
  (finalout :initarg finalout :initform "") ;; what is to pbe played at end?.
  (rescale :initarg rescale :initform NIL)  ;; rescale before play? (output is floating-point file)
  (syntax :initform (find-syntax ':cmix))))

(defmethod open-event-stream ((syntax cmix) &rest args)
  (apply #'open-event-stream (find-class 'cmix-score-file) args))

(defmethod initialize-stream-for-processing ((stream cmix-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "/* Stella output ~A */~%~%  " (cm::date-string)) 
    (when (slot-boundp stream 'header)
      (format file "~A~%" (slot-value stream 'header)))))
#|
;; No need for cmix..
(defmethod deinitialize-stream-for-processing ((stream cmix-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "ter ~A;~%" (if (eql (slot-value stream 'revtail) 0.0) 	
					(slot-value stream 'last-time)
 	(+ (slot-value stream 'revtail) (slot-value stream 'last-time))))))
|#

(defmethod play-using-syntax ((syntax cmix) file &rest args)
	(declare (ignore args))
  #+mcl (format t "Sorry, can't play cmix from Lisp on the Mac.~%")
  #-mcl	(let* ((orc  (slot-value (find-stream file) 'orchestra))
		(ofile (slot-value (find-stream file) 'finalout))
		(command (format nil "~A < ~A " orc 
					(namestring (truename file))))
		(plcmd (if (null (slot-value (find-stream file) 'rescale))
			(format nil "~Aunix/playsnd ~A" 
					*common-music-directory* ofile)
			(format nil "~Aunix/rescalesnd ~A;~Aunix/playsnd ~A.short"
			 *common-music-directory* ofile))))
(format t "Executing cmix with:~%Output=~A ~%Instruments=~A ~%Score=~A~%"
				 ofile orc (namestring (truename file))) 
		(cm::shell command)
		(tell-user "Playing ~A ." ofile)
		(cm::shell plcmd)))

;;; Here are the standard cmix objects.....
;;; input
(defobject input (note)
((name :initform "input") filename (unitnumber :initform 0) (rhythm :initform 0.0)))

(defmethod print-object ((object input) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :default +slot-unset+)
      name time filename unitnumber)))

(defmethod write-event ((object input) (stream cmix-score-file)) 
  (formatting-slots (object (slot-value stream 'stream) :postamble ")" :preamble "input(" :delimiter #\,  :eol t)
     (filename :printer prin1) unitnumber))

(defmethod write-event :after ((object input) (stream cmix-score-file))
(if (member (slot-value object 'unitnumber) (slot-value stream 'openfiles)) NIL
		(push (slot-value object 'unitnumber) (slot-value stream 'openfiles))) ;else
(if  (> (length  (slot-value stream 'openfiles)) 5) (error "Too Many i/o files open in cmix score!")))

;;; output

 (defobject output (note)
((name :initform "output") filename (unitnumber :initform 1) (rhythm :initform 0.0)))

(defmethod print-object ((object output) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :default +slot-unset+)
      name filename unitnumber)))

(defmethod write-event ((object output) (stream cmix-score-file)) 
  (formatting-slots (object (slot-value stream 'stream) :postamble ")" :preamble "output(" :delimiter #\, :eol t)
     (filename :printer prin1) unitnumber))

(defmethod write-event :after ((object output) (stream cmix-score-file))
(if (member (slot-value object 'unitnumber) (slot-value stream 'openfiles)) NIL
		(push (slot-value object 'unitnumber) (slot-value stream 'openfiles))) ;else
(if  (> (length  (slot-value stream 'openfiles)) 5) (error "Too Many i/o files open in cmix score!")))

;; setline
(defobject setline (note)
((name :initform "setline") break-points (rhythm :initform 0.0) ))

(defmethod print-object ((object setline) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :default +slot-unset+) name break-points)))

(defmethod write-event ((object setline) (stream cmix-score-file)) 
(let ((strm (slot-value stream 'stream))(bp (slot-value object 'break-points)))
(format strm "setline(")
(loop for i in (butlast bp)
do (format strm "~A," i)
finally (progn (format strm "~A" (car (last bp))) (format strm ")~%")))))

;;; mix - first, see cmix docs, then:
;;; outskip is time! end is dur! channel0 to channel 0, chan 1 to 1, 
;;; no chans3&4 is default
(defobject mix (note) 
((name :initform "mix") inskip outskip dur amp (ch0loc :initform 0) (ch1loc :initform 1) (ch2loc :initform -1)(ch3loc :initform -1) (rhythm :initform 0.0)))

(defmethod print-object ((object mix) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :default +slot-unset+) name inskip outskip dur amp ch0loc ch1loc)))

(defmethod write-event ((object mix) (stream cmix-score-file))
 (if  (slot-boundp object 'outskip) () (setf (slot-value object 'outskip) (slot-value object 'time)))
  (formatting-slots (object (slot-value stream 'stream) :postamble ")" :preamble "mix(" :delimiter #\, :eol t)
     inskip outskip dur amp ch0loc ch1loc ch2loc ch3loc))

;;; trans 
(defobject trans (note) 
((name :initform "trans") inskip outskip dur pitch_trans amp (rhythm :initform 0.0)))

(defmethod print-object ((object trans) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :default +slot-unset+) name inskip outskip dur pitch_trans amp )))

(defmethod write-event ((object trans) (stream cmix-score-file)) 
(if  (slot-boundp object 'outskip) () (setf (slot-value object 'outskip) (slot-value object 'time)))
  (formatting-slots (object (slot-value stream 'stream) :postamble ")" :preamble "trans(" :delimiter #\, :eol t)
     inskip outskip dur pitch_trans amp))

